#!/usr/bin/env python3

# Copyright (C) 2017-2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Include a commitment inside an elliptic curve DSA/SSA signature.

Let c be the commitment value and R a curve point, then

    e = hash(R||c)

is a commitment operation.

When signing, an ephemeral secret key k is generated and its
corresponding curve point R = kG is used. Here, instead of
using (k, R), compute the commitment to c

    e = hash(R||c),

tweak k with e and consequently substistute R with W = (k+e)G = R+eG,
the proceed signing in the standard way, using (k+e, W).

When the committer/signer will reveal R and c,
the verifier will check that

    W.x = (R+eG).x

with e = hash(R||c)) and W.x being known from the signature.
"""

from typing import Optional, Tuple

from .curve import Curve, mult, Point
from .utils import int_from_bits, point_from_octets, octets_from_point, HashF
from .rfc6979 import rfc6979
from . import dsa
from . import ssa

# commitment receipt
Receipt = Tuple[int, Point]


def _tweak(ec: Curve, hf: HashF, c: bytes, k: int) -> Tuple[Point, int]:
    """Tweak kG with hash(kG||c).

    Return:
    - the point kG to tweak
    - tweaked private key k + hash(kG||c)
    """
    R = mult(ec, k)
    e = hf(octets_from_point(ec, R, True) + c).digest()
    e = int.from_bytes(e, 'big')
    return R, (e + k) % ec.n


def ecdsa_commit_sign(ec: Curve, hf: HashF, c: bytes, m: bytes, prvkey: int,
                      k: Optional[int] = None) -> Tuple[dsa.ECDS, Receipt]:
    """Include a commitment c inside an ECDSA signature."""

    if k is None:
        k = rfc6979(ec, hf, hf(m).digest(), prvkey)

    ch = hf(c).digest()

    # commit
    R, new_k = _tweak(ec, hf, ch, k)
    # sign
    sig = dsa.sign(ec, hf, m, prvkey, new_k)
    # commit receipt
    receipt = sig[0], R
    return sig, receipt


def ecssa_commit_sign(ec: Curve, hf: HashF, c: bytes, m: bytes, prvkey: int,
                      k: Optional[int] = None) -> Tuple[ssa.ECSS, Receipt]:
    """Include a commitment c inside an ECSSA signature."""

    if k is None:
        k = rfc6979(ec, hf, m, prvkey)

    ch = hf(c).digest()

    # commit
    R, new_k = _tweak(ec, hf, ch, k)
    # sign
    sig = ssa.sign(ec, hf, m, prvkey, new_k)
    # commit receipt
    receipt = sig[0], R
    return sig, receipt

# FIXME: have create_commit instead of commit_sign


def verify_commit(ec: Curve, hf: HashF, c: bytes, receipt: Receipt) -> bool:
    """Open the commitment c inside an EC DSA/SSA signature."""

    # FIXME: verify the signature

    w, R = receipt
    # w in [1..n-1] dsa
    # w in [1..p-1] ssa
    # different verify functions?
    
    # verify R is a good point?    

    ch = hf(c).digest()
    e = hf(octets_from_point(ec, R, True) + ch).digest()
    e = int_from_bits(ec, e)
    W = ec.add(R, mult(ec, e))
    # different verify functions?
    # return w == W[0] # ECSS
    return w == W[0] % ec.n  # ECDS, FIXME: ECSSA



#!/usr/bin/env python3

# Copyright (C) 2017-2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Assorted conversion utilities.

Most conversions from SEC 1 v.2 2.3 are included.

https://www.secg.org/sec1-v2.pdf
"""

from typing import Union, Callable, Any
from hashlib import sha256, new

from .curve import Curve, Point

HashF = Callable[[Any], Any]

# bytes or hex string
Octets = Union[str, bytes]


def point_from_octets(ec: Curve, o: Octets) -> Point:
    """Return a tuple (Px, Py) that belongs to the curve.

    Return a tuple (Px, Py) that belongs to the curve according to
    SEC 1 v.2, section 2.3.4.
    """

    if isinstance(o, str):
        o = bytes.fromhex(o)

    bsize = len(o) # bytes
    if bsize == 1 and o[0] == 0x00:     # infinity point
        return Point()

    if bsize == ec.psize+1:             # compressed point
        if o[0] not in (0x02, 0x03):
            m = f"{ec.psize+1} bytes, but not a compressed point"
            raise ValueError(m)
        Px = int.from_bytes(o[1:], 'big')
        try:
            Py = ec.y_odd(Px, o[0] % 2)  # also check Px validity
            return Point(Px, Py)
        except:
            raise ValueError("point not on curve")
    else:                               # uncompressed point
        if bsize != 2*ec.psize+1:
            m = f"wrong byte-size ({bsize}) for a point: it "
            m += f"should have be {ec.psize+1} or {2*ec.psize+1}"
            raise ValueError(m)
        if o[0] != 0x04:
            raise ValueError("not an uncompressed point")
        Px = int.from_bytes(o[1:ec.psize+1], 'big')
        P = Point(Px, int.from_bytes(o[ec.psize+1:], 'big'))
        if ec.is_on_curve(P):
            return P
        else:
            raise ValueError("point not on curve")


def octets_from_point(ec: Curve, Q: Point, compressed: bool) -> bytes:
    """Return a point as compressed/uncompressed octet sequence.
    
    Return a point as compressed (0x02, 0x03) or uncompressed (0x04)
    octet sequence, according to SEC 1 v.2, section 2.3.3.
    """

    # check that Q is a point and that is on curve
    ec.require_on_curve(Q)

    if Q[1] == 0:  # infinity point in affine coordinates
        return b'\x00'

    bPx = Q[0].to_bytes(ec.psize, byteorder='big')
    if compressed:
        return (b'\x03' if (Q[1] & 1) else b'\x02') + bPx

    return b'\x04' + bPx + Q[1].to_bytes(ec.psize, byteorder='big')


def int_from_octets(o: Octets) -> int:
    """Return an integer from an octet sequence (bytes or hex string).
    
    Return an integer from an octet sequence (bytes or hex string)
    according to SEC 1 v.2, section 2.3.8.
    """
    if isinstance(o, str):  # hex string
        o = bytes.fromhex(o)
    return int.from_bytes(o, 'big')


def octets_from_int(i: int, bytesize: int) -> bytes:
    """Return an octet sequence from an integer.
    
    Return an octet sequence from an integer
    according to SEC 1 v.2, section 2.3.7.
    """

    return i.to_bytes(bytesize, 'big')


def int_from_bits(ec: Curve, o: Octets) -> int:
    """Return the leftmost nlen bits reduced modulo n.

    Take as input a sequence of blen bits and calculate a
    non-negative integer i that is less than 2^nlen according to
    SEC 1 v.2 section 4.1.3 (5). Further,
    reduce i module n to ensure that 0 < i < n.

    int_from_bits is used during signature generation and verification
    in ECDSA and ECSSA to transform a hash value (computed over the
    input message) into an integer modulo n.

    Note that int_from_bits is not the reverse of octets_from_int, even
    for input sequences of length nlen: octets_from_int will add some
    bits on the left, while int_from_bits will discard some bits on the
    right. octets_from_int is the reverse of int_from_bits only when
    nlen is a multiple of 8 and bit sequences already have length nlen.
    See https://tools.ietf.org/html/rfc6979#section-2.3.5.
    """
    i = _int_from_bits(ec, o)
    return i % ec.n  # might be implemented as difference


def _int_from_bits(ec: Curve, o: Octets) -> int:
    """Return the leftmost nlen bits.

    Take as input a sequence of blen bits and calculate a
    non-negative integer i that is less than 2^nlen according to
    SEC 1 v.2 section 4.1.3 (5).  Note that an additional
    reduction modulo n would be required to ensure that 0 < i < n.
    """
    i = int_from_octets(o)

    blen = len(o) * 8  # bits
    n = (blen - ec.nlen) if blen >= ec.nlen else 0
    return i >> n


def h160(o: Octets) -> bytes:
    """Return RIPEMD160(SHA256) of an octet sequence."""

    if isinstance(o, str):  # hex string
        o = bytes.fromhex(o)

    t = sha256(o).digest()
    return new('ripemd160', t).digest()


def double_sha256(o: Octets) -> bytes:
    """Return SHA256(SHA256()) of an octet sequence."""

    if isinstance(o, str):
        o = bytes.fromhex(o)

    return sha256(sha256(o).digest()).digest()



#!/usr/bin/env python3

# Copyright (C) 2017-2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Base58 encoding and decoding functions.

Binary-to-text encoding schemes are designed to transport binary data across
channels that are designed to deal with textual data. In Bitcoin they are mostly
used to represent large integers as alphanumeric text.

Base58 is similar to Base64, which uses 10 digits, 26 lowercase characters,
26 uppercase characters, '+' (plus sign), and '/' (forward slash).
Base58 omits the similar-looking letters
0 (zero), O (capital o), I (capital i), and l (lower case L)
to avoid ambiguity when printed; moreover, it removes '+' and '/'
so that a double-click does select the whole string.

This implementation of Base58 and Base58Check is originally from
https://github.com/keis/base58, with the following modifications:

* type annotated python3
* using native python3 int.from_bytes() and i.to_bytes()
* added optional check on output size for _decode() and decode_check()

Input can be bytes or a string that will be encoded to bytes (after
being stripped of leading/trailing white spaces)
"""

from hashlib import sha256
from typing import Union, Optional

from .utils import double_sha256

# used digits
__digits = b'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
__base = len(__digits)


def _str_to_bytes(v: Union[str, bytes]) -> bytes:
    """Encode string to bytes, stipping leading/trailing white spaces"""

    if isinstance(v, str):
        v = v.strip()
        v = v.encode()

    return v


def _encode_from_int(i: int) -> bytes:
    """Encode an integer using Base58."""

    if i == 0:
        return __digits[0:1]

    result = b""
    while i:
        i, idx = divmod(i, __base)
        result = __digits[idx:idx+1] + result

    return result


def _encode(v: Union[str, bytes]) -> bytes:
    """Encode bytes or string using Base58."""

    v = _str_to_bytes(v)

    # preserve leading-0s
    # leading-0s become base58 leading-1s
    nPad = len(v)
    v = v.lstrip(b'\0')
    vlen = len(v)
    nPad -= vlen
    result = __digits[0:1] * nPad

    if vlen:
        i = int.from_bytes(v, 'big')
        result += _encode_from_int(i)

    return result


def encode_check(v: Union[str, bytes]) -> bytes:
    """Encode bytes or string using checksummed Base58."""

    v = _str_to_bytes(v)

    digest = double_sha256(v)
    return _encode(v + digest[:4])


def _decode_to_int(v: Union[str, bytes]) -> int:
    """Decode Base58 encoded bytes or string as integer."""

    v = _str_to_bytes(v)

    i = 0
    for char in v:
        i *= __base
        i += __digits.index(char)
    return i


def _decode(v: Union[str, bytes],
            output_size: Optional[int] = None) -> bytes:
    """Decode Base58 encoded bytes or string.
    
    Decode Base58 encoded bytes or string,
    optionally ensuring required output size.
    """


    v = _str_to_bytes(v)

    # preserve leading-0s
    # base58 leading-1s become leading-0s
    nPad = len(v)
    v = v.lstrip(__digits[0:1])
    vlen = len(v)
    nPad -= vlen
    result = b'\0' * nPad

    if vlen:
        i = _decode_to_int(v)
        nbytes = (i.bit_length() + 7) // 8
        result = result + i.to_bytes(nbytes, 'big')

    if output_size is not None and len(result) != output_size:
        m = "Invalid decoded size: "
        m += f"{len(result)} bytes instead of {output_size}"
        raise ValueError(m)

    return result


def decode_check(v: Union[str, bytes],
                 output_size: Optional[int] = None) -> bytes:
    """Decode Base58 encoded bytes or string, verifying checksum.
    
    Decode Base58 encoded bytes or string, verifying checksum and
    optionally ensuring required output size.
    """

    if output_size is not None:
        output_size += 4

    result = _decode(v, output_size)
    result, checksum = result[:-4], result[-4:]

    digest = double_sha256(result)
    if checksum != digest[:4]:
        m = f"Invalid checksum: '{checksum}' instead of '{digest[:4]}'"
        raise ValueError(m)

    return result



#!/usr/bin/env python3

# Copyright (C) 2017-2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""BIP32 Hierarchical Deterministic Wallet functions.

A deterministic wallet is a hash-chain of private/public key pairs that
derives from a single root, which is the only element requiring backup.
Moreover, there are schemes where public keys can be calculated without
accessing private keys.

A hierarchical deterministic wallet is a tree of multiple hash-chains,
derived from a single root, allowing for selective sharing of keypair
chains.

Here, the HD wallet is implemented according to BIP32 bitcoin standard
https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki.
"""

from hmac import HMAC
from hashlib import sha512
from typing import Union, Optional, Sequence, List, Tuple

from . import base58 
from .curve import mult
from .curves import secp256k1 as ec
from .utils import Octets, point_from_octets, octets_from_point, \
    int_from_octets, h160
from .wifaddress import p2pkh_address

# Bitcoin core uses the m/0h (core) BIP32 derivation path
# with xprv/xpub and tprv/tpub Base58 encoding

# VERSION BYTES =      4 bytes
# m/44h/0h  p2pkh or p2sh
MAIN_xprv = b'\x04\x88\xAD\xE4'; MAIN_xpub = b'\x04\x88\xB2\x1E'
# m/44h/1h  p2pkh or p2sh
TEST_tprv = b'\x04\x35\x83\x94'; TEST_tpub = b'\x04\x35\x87\xCF'
# m/49h/0h  p2sh-segwit p2wpkh-p2sh
MAIN_yprv = b'\x04\x9D\x78\x78'; MAIN_ypub = b'\x04\x9D\x7C\xB2'
# m/49h/1h  p2sh-segwit p2wpkh-p2sh
TEST_uprv = b'\x04\x4A\x4E\x28'; TEST_upub = b'\x04\x4A\x52\x62'
# m/84h/0h  native segwit P2WPKH
MAIN_zprv = b'\x04\xB2\x43\x0C'; MAIN_zpub = b'\x04\xB2\x47\x46'
# m/84h/1h  native segwit P2WPKH
TEST_vprv = b'\x04\x5F\x18\xBC'; TEST_vpub = b'\x04\x5F\x1C\xF6'
#   ---     p2sh-segwit multi-sig p2wpkh-p2sh
MAIN_Yprv = b'\x02\x95\xB0\x05'; MAIN_Ypub = b'\x02\x95\xB4\x3F'
TEST_Uprv = b'\x02\x42\x85\xB5'; TEST_Upub = b'\x02\x42\x89\xEF'
#   ---     native segwit multi-sig p2wpkh
MAIN_Zprv = b'\x02\xAA\x7A\x99'; MAIN_Zpub = b'\x02\xAA\x7E\xD3'
TEST_Vprv = b'\x02\x57\x50\x48'; TEST_Vpub = b'\x02\x57\x54\x83'


PRV_VERSION = [
    MAIN_xprv, MAIN_yprv, MAIN_zprv, MAIN_Yprv, MAIN_Zprv,
    TEST_tprv, TEST_uprv, TEST_vprv, TEST_Uprv, TEST_Vprv]
PUB_VERSION = [
    MAIN_xpub, MAIN_ypub, MAIN_zpub, MAIN_Ypub, MAIN_Zpub,
    TEST_tpub, TEST_upub, TEST_vpub, TEST_Upub, TEST_Vpub]

# [  : 4] version
# [ 4: 5] depth
# [ 5: 9] parent _pubkey_ fingerprint
# [ 9:13] child index
# [13:45] chain code
# [45:78] key (private/public)

def rootxprv_from_seed(seed: Octets, version: Octets) -> bytes:
    """derive the BIP32 root master extended private key from the seed"""

    if isinstance(version, str):  # hex string
        version = bytes.fromhex(version)
    if version not in PRV_VERSION:
        msg = f"invalid private version ({version})"
        raise ValueError(msg)

    # serialization data
    rootxprv = version                            # version
    rootxprv += b'\x00'                           # depth
    rootxprv += b'\x00\x00\x00\x00'               # parent pubkey fingerprint
    rootxprv += b'\x00\x00\x00\x00'               # child index

    # actual extended key (key + chain code) derivation
    if isinstance(seed, str):  # hex string
        seed = bytes.fromhex(seed)
    hd = HMAC(b"Bitcoin seed", seed, sha512).digest()
    rootprv = int_from_octets(hd[:32])
    rootxprv += hd[32:]                                # chain code
    rootxprv += b'\x00' + rootprv.to_bytes(32, 'big')  # private key

    return base58.encode_check(rootxprv)


def xpub_from_xprv(xprv: Octets) -> bytes:
    """Neutered Derivation (ND)

    Computation of the extended public key corresponding to an extended
    private key (“neutered” as it removes the ability to sign transactions)
    """

    xprv = base58.decode_check(xprv, 78)
    if xprv[45] != 0:
        raise ValueError("extended key is not a private one")

    i = PRV_VERSION.index(xprv[:4])

    # serialization data
    xpub = PUB_VERSION[i]                      # version
    # unchanged serialization data
    xpub += xprv[4: 5]                         # depth
    xpub += xprv[5: 9]                         # parent pubkey fingerprint
    xpub += xprv[9:13]                         # child index
    xpub += xprv[13:45]                        # chain code

    p = int_from_octets(xprv[46:])
    P = mult(ec, p)
    xpub += octets_from_point(ec, P, True)          # public key
    return base58.encode_check(xpub)


def ckd(xparentkey: Octets, index: Union[Octets, int]) -> bytes:
    """Child Key Derivation (CDK)

    Key derivation is normal if the extended parent key is public or
    child_index is less than 0x80000000.

    Key derivation is hardened if the extended parent key is private and
    child_index is not less than 0x80000000.
    """

    if isinstance(index, int):
        index = index.to_bytes(4, 'big')
    elif isinstance(index, str):  # hex string
        index = bytes.fromhex(index)

    if len(index) != 4:
        raise ValueError(f"a 4 bytes int is required, not {len(index)}")

    xparent = base58.decode_check(xparentkey, 78)

    version = xparent[:4]

    # serialization data
    xkey = version                               # version
    xkey += (xparent[4] + 1).to_bytes(1, 'big')  # (increased) depth

    if version in PUB_VERSION:
        if xparent[45] not in (2, 3):  # not a compressed public key
            raise ValueError("(pubversion/prvkey) mismatch in extended parent key")
        Parent_bytes = xparent[45:]
        Parent = point_from_octets(ec, Parent_bytes)
        xkey += h160(Parent_bytes)[:4]           # parent pubkey fingerprint
        if index[0] >= 0x80:
            raise ValueError("no private/hardened derivation from pubkey")
        xkey += index                            # child index
        parent_chain_code = xparent[13:45]       # normal derivation
        # actual extended key (key + chain code) derivation
        h = HMAC(parent_chain_code, Parent_bytes + index, sha512).digest()
        offset = int.from_bytes(h[:32], 'big')
        Offset = mult(ec, offset)
        Child = ec.add(Parent, Offset)
        Child_bytes = octets_from_point(ec, Child, True)
        xkey += h[32:]                           # chain code
        xkey += Child_bytes                      # public key
    elif version in PRV_VERSION:
        if xparent[45] != 0:    # not a private key
            raise ValueError("(prvversion/pubkey) mismatch in extended parent key")
        parent = int.from_bytes(xparent[46:], 'big')
        Parent = mult(ec, parent)
        Parent_bytes = octets_from_point(ec, Parent, True)
        xkey += h160(Parent_bytes)[:4]           # parent pubkey fingerprint
        xkey += index                            # child index
        # actual extended key (key + chain code) derivation
        parent_chain_code = xparent[13:45]
        if (index[0] < 0x80):                    # normal derivation
            h = HMAC(parent_chain_code, Parent_bytes + index, sha512).digest()
        else:                                    # hardened derivation
            h = HMAC(parent_chain_code, xparent[45:] + index, sha512).digest()
        offset = int.from_bytes(h[:32], 'big')
        child = (parent + offset) % ec.n
        child_bytes = b'\x00' + child.to_bytes(32, 'big')
        xkey += h[32:]                           # chain code
        xkey += child_bytes                      # private key
    else:
        raise ValueError("invalid extended key version")

    return base58.encode_check(xkey)


def indexes_from_path(path: str) -> Tuple[Sequence[int], bool]:
    """Extract derivation indexes from a derivation path like
       "m/44'/0'/1'/0/10" (absolute) or "./0/10" (relative)
    """

    steps = path.split('/')
    if steps[0] == 'm':
        absolute = True
    elif steps[0] == '.':
        absolute = False
    else:
        raise ValueError(f'Invalid derivation path: {path}')

    indexes: List[int] = list()
    for step in steps[1:]:
        hardened = False
        if step[-1] in ("'", "H", "h"):
            hardened = True
            step = step[:-1]
        index = int(step)
        index += 0x80000000 if hardened else 0
        indexes.append(index)

    return indexes, absolute


def derive(xkey: Octets, path: Union[str, Sequence[int]]) -> bytes:
    """derive an extended key according to path like
       "m/44'/0'/1'/0/10" (absolute) or "./0/10" (relative)
    """

    if isinstance(path, str):
        indexes, absolute = indexes_from_path(path)
        if absolute:
            decoded = base58.decode_check(xkey, 78)
            t = b'\x00'*9
            if decoded[4:13] != t:
                msg = "Absolute derivation path for non-root master key"
                raise ValueError(msg)
    else:
        indexes = path

    for index in indexes:
        xkey = ckd(xkey, index)

    return xkey


def p2pkh_address_from_xpub(xpub: Octets, version: Optional[Octets] = None) -> bytes:
    xpub = base58.decode_check(xpub, 78)
    if xpub[45] not in (2, 3):
        raise ValueError("extended key is not a public one")

    if version is None:
        xversion = xpub[:4]
        if xversion == MAIN_xpub:
            version = b'\x00'          # 1
        elif xversion == TEST_tpub:
            version = b'\x6F'          # m or n
        else:
            raise ValueError(f"xkey is not of p2pkh type (xpub/tpub)")

    P = point_from_octets(ec, xpub[45:])
    return p2pkh_address(P, True, version)


def crack(parent_xpub: Octets, child_xprv: Octets) -> bytes:
    parent_xpub = base58.decode_check(parent_xpub, 78)
    if parent_xpub[45] not in (2, 3):
        raise ValueError("extended parent key is not a public one")

    child_xprv = base58.decode_check(child_xprv, 78)
    if child_xprv[45] != 0:
         raise ValueError("extended child key is not a private one")

    # check depth
    if child_xprv[4] != parent_xpub[4] + 1:
         raise ValueError("wrong child/parent depth relation")

    # check fingerprint
    Parent_bytes = parent_xpub[45:]
    if child_xprv[5: 9] != h160(Parent_bytes)[:4]:
        raise ValueError("not a child for the provided parent")

    # check normal derivation
    child_index = child_xprv[9:13]
    if child_index[0] >= 0x80:
        raise ValueError("hardened derivation")

    parent_xprv = child_xprv[: 4]     # version
    parent_xprv += parent_xpub[4: 5]  # depth
    parent_xprv += parent_xpub[5: 9]  # parent pubkey fingerprint
    parent_xprv += parent_xpub[9:13]  # child index

    parent_chain_code = parent_xpub[13:45]
    parent_xprv += parent_chain_code  # chain code

    h = HMAC(parent_chain_code, Parent_bytes + child_index, sha512).digest()
    offset = int.from_bytes(h[:32], 'big')
    child = int.from_bytes(child_xprv[46:], 'big')
    parent = (child - offset) % ec.n
    parent_bytes = b'\x00' + parent.to_bytes(32, 'big')
    parent_xprv += parent_bytes        # private key

    return base58.encode_check(parent_xprv)


def child_index(xkey: Octets) -> bytes:
    xkey = base58.decode_check(xkey, 78)
    if xkey[4] == 0:
        raise ValueError("master key provided")
    return xkey[9:13]

def fingerprint(xkey: Octets) -> bytes:
    key = base58.decode_check(xkey, 78)
    version = key[:4]

    if key[45] == 0:
        if version in PRV_VERSION:
            xkey = xpub_from_xprv(xkey)
            key = base58.decode_check(xkey, 78)
        else:
            raise ValueError("(pubversion/prvkey) mismatch in extended parent key")
    elif key[45] in (2, 3):
        if version not in PUB_VERSION:
            raise ValueError("(prvversion/pubkey) mismatch in extended parent key")
    else:
        ValueError("not a valid extended key")

    return h160(key[45:])[:4]



#!/usr/bin/env python3

# Copyright (C) 2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Bitcoin (P2PKH) address-based compact signature for messages.

For message signatures, Bitcoin wallets use a P2PKH address-based scheme
with a compact 65 bytes custom signature encoding.
As it is the case for all digital signature scheme, this scheme actually
works with keys, not addresses: it uses a P2PKH address to uniquely
identify a private/public keypair.
This signature proves the control of the private key corresponding to
a given address and, consequently, of the associated bitcoins (if any).
The signature goes along with its address: public key recovery is used
at verification time, i.e. given a message, the public key
that would have created that signature is found and compared with
the provided address.

Note that in the Bitcoin protocol this compact 65 bytes signature
encoding is only used for messages: for transactions Bitcoin uses DER
encoding instead, resulting in 71 bytes average signature.

At signing time a wallet infrastructure is required to access the
private key corresponding to the given address; alternatively
the private key must be provided explicitly.
For a given message, the ECDSA signature operates on the hash of the
*magic* "Bitcoin Signed Message" prefix concatenated to the actual
message; this prefix manipulation avoids the plain signature of a
possibly deceiving message.
The resulting (r, s) signature is serialized as
[1 byte][r][s], where the first byte is a recovery flag used
during signature verification to discriminate among recovered
public keys and to manage address compression.
Explicitly, the recovery flag value is:

    27 + (IF compressed THEN 4 ELSE 0) + key_id

where:

- 27 identify a P2PKH address (Electrum also supports Segwit P2WPKH-P2SH
  and P2WPKH, but not according to the BIP137 specifications;
  anyway this module and bitcoin core do not support them yet)
- compressed indicates if the address is the hash of the compressed
  public key representation (Segwit is always compressed)
- key_id is the index in the [0, 3] range identifying which of the
  recovered public keys is the one associated to the address;
  it is stored in the least significant 2 bits of the header

+-----------+--------+--------------------+
| rec. flag | key_id |    address type    |
+===========+========+====================+
|     27    |    0   | P2PKH uncompressed |
+-----------+--------+--------------------+
|     28    |    1   | P2PKH uncompressed |
+-----------+--------+--------------------+
|     29    |    2   | P2PKH uncompressed |
+-----------+--------+--------------------+
|     30    |    3   | P2PKH uncompressed |
+-----------+--------+--------------------+
|     31    |    0   | P2PKH compressed   |
+-----------+--------+--------------------+
|     32    |    1   | P2PKH compressed   |
+-----------+--------+--------------------+
|     33    |    2   | P2PKH compressed   |
+-----------+--------+--------------------+
|     34    |    3   | P2PKH compressed   |
+-----------+--------+--------------------+
|     35    |    0   | P2WPKH-P2SH        |
+-----------+--------+--------------------+
|     36    |    1   | P2WPKH-P2SH        |
+-----------+--------+--------------------+
|     37    |    2   | P2WPKH-P2SH        |
+-----------+--------+--------------------+
|     38    |    3   | P2WPKH-P2SH        |
+-----------+--------+--------------------+
|     39    |    0   | P2WPKH (bech32)    |
+-----------+--------+--------------------+
|     40    |    1   | P2WPKH (bech32)    |
+-----------+--------+--------------------+
|     41    |    2   | P2WPKH (bech32)    |
+-----------+--------+--------------------+
|     42    |    3   | P2WPKH (bech32)    |
+-----------+--------+--------------------+

Finally, the serialized signature can be base64-encoded to transport it
across channels that are designed to deal with textual data.
Base64-encoding uses 10 digits, 26 lowercase characters, 26 uppercase
characters, '+' (plus sign), and '/' (forward slash); equal sign '=' is
used as 65th character pad, a complement in the final process of
message encoding.

Warning: one should never sign a vague statement that could be reused
out of the context it was intended for. E.g. always include at least

- your name (nickname, customer id, email, etc.)
- date and time
- who the message is intended for (name, business name, email, etc.)
- specific purpose of the message

https://bitcoin.stackexchange.com/questions/10759/how-does-the-signature-verification-feature-in-bitcoin-qt-work-without-a-public

https://bitcoin.stackexchange.com/questions/12554/why-the-signature-is-always-65-13232-bytes-long

https://bitcoin.stackexchange.com/questions/34135/what-is-the-strmessagemagic-good-for

https://bitcoin.stackexchange.com/questions/36838/why-does-the-standard-bitcoin-message-signature-include-a-message-prefix

https://bitcoin.stackexchange.com/questions/68844/explicit-message-length-in-bitcoin-signed-message

https://github.com/bitcoinjs/bitcoinjs-lib/blob/1079bf95c1095f7fb018f6e4757277d83b7b9d07/src/message.js#L13

https://bitcointalk.org/index.php?topic=6428

https://bitcointalk.org/index.php?topic=6430

https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106?newreg=670c5855241d4340af0cbbc960fd2dc3

https://github.com/bitcoin/bitcoin/pull/524

https://www.reddit.com/r/Bitcoin/comments/bgcgs2/can_bitcoin_core_0171_sign_message_from_segwit/

https://github.com/bitcoin/bips/blob/master/bip-0137.mediawiki

https://github.com/brianddk/bips/blob/legacysignverify/bip-0xyz.mediawiki

"""

import base64
from hashlib import sha256 as hf
from typing import Tuple, Union

from .base58 import _str_to_bytes
from .curve import mult
from .curves import secp256k1 as ec
from .wifaddress import p2pkh_address, h160_from_pubkey, _h160_from_address
from . import dsa

# TODO: support msg as bytes
# TODO: add testnet / regtest / litecoin signature
# TODO: add small wallet (address <-> private key) infrastructure
# TODO:                           then also add sign(address, msg)
# TODO: decouple serialization from address-based signature
# TODO: add test vectors from P. Todd's library
# TODO: report Electrum bug
# TODO: generalize to other curves and hash functions
# TODO: implement P2WPKH-P2SH and P2WPKH

def _magic_hash(msg: str) -> bytes:
    # Electrum does strip leading and trailing spaces;
    # bitcoin core does not
    # msg = msg.strip()
    m = hf()
    prefix = b'\x18Bitcoin Signed Message:\n'
    m.update(prefix)
    message = chr(len(msg)) + msg
    m.update(message.encode())
    return m.digest()

def sign(msg: str, prvkey: int, compressed: bool) -> Tuple[str, str]:
    """Generate message signature for a given P2PKH address."""

    pubkey = mult(ec, prvkey)
    address = p2pkh_address(pubkey, compressed, b'\x00')

    magic_msg = _magic_hash(msg)
    sig = dsa.sign(ec, hf, magic_msg, prvkey)

    pubkeys = dsa.pubkey_recovery(ec, hf, magic_msg, sig)
    sig = sig[0].to_bytes(32, 'big') + sig[1].to_bytes(32, 'big')
    for i in range(len(pubkeys)):
        if pubkeys[i] == pubkey:
            rf = 27 + i
            if compressed:
                rf += 4
            return address, base64.b64encode(bytes([rf]) + sig)

    # the following line should never be executed
    raise ValueError("Public key not recovered")

def verify(msg: str, addr: Union[str, bytes], sig: Union[str, bytes]) -> bool:
    """Verify message signature for a given P2PKH address."""

    # try/except wrapper for the Errors raised by _verify
    try:
        return _verify(msg, addr, sig)
    except Exception:
        return False


def _verify(msg: str, addr: Union[str, bytes], sig: Union[str, bytes]) -> bool:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    sig = base64.b64decode(sig)
    if len(sig) != 65:
        raise ValueError(f"Wrong encoding length: {len(sig)} instead of 65")

    r = int.from_bytes(sig[1:33], 'big')
    s = int.from_bytes(sig[33:], 'big')
    magic_msg = _magic_hash(msg)
    pubkeys = dsa.pubkey_recovery(ec, hf, magic_msg, (r, s))

    # almost any sig/msg pair recovers (a pubkey and) an addr:
    # signature is valid only if the provided addr is matched
    addr = _str_to_bytes(addr)
    rf = int.from_bytes(sig[0:1], 'big')
    if rf < 27 or rf > 42:
        raise ValueError(f"Unknown recovery flag: {rf}")
    elif rf > 38 or addr.startswith(b'bc1'):
        raise ValueError("P2WPKH bech32 address not supported yet")
    elif rf > 34 or addr.startswith(b'3'):
        raise ValueError("P2WPKH-P2SH address not supported yet")
    else:
        # verify that input addr is a valid P2PKH addr
        h160 = _h160_from_address(addr)
        # i selects which key is recovered
        i = (rf - 27) & 3
        if rf < 31:
            return h160 == h160_from_pubkey(pubkeys[i], False)
        else:
            return h160 == h160_from_pubkey(pubkeys[i], True)



#!/usr/bin/env python3

# Copyright (C) 2017-2019 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Class for converting entropy from/to mnemonic sentence.

*Entropy* must be represented as binary 0/1 string.
"""

import math
from hashlib import pbkdf2_hmac
from typing import List

from .entropy import Entropy
from .wordlists import _wordlists

Mnemonic = str

def indexes_from_entropy(entropy: Entropy, lang: str) -> List[int]:
    """Return the word-list indexes for a given entropy.
    
    Return the list of integer indexes into a language word-list
    for a given entropy.

    Entropy must be represented as binary 0/1 string; leading zeros
    are not considered redundant padding.
    """

    if type(entropy) != str:
        m = "entropy must be binary string, "
        m += f"not '{type(entropy).__name__}'"
        raise TypeError(m)

    bits = len(entropy)
    int_entropy = int(entropy, 2)
    n = _wordlists.language_length(lang)
    indexes = []
    while int_entropy:
        int_entropy, index = divmod(int_entropy, n)
        indexes.append(index)

    # do not lose leading zeros entropy
    bpw = _wordlists.bits_per_word(lang)
    nwords = math.ceil(bits/bpw)
    while len(indexes) < nwords:
        indexes.append(0)

    return list(reversed(indexes))

def mnemonic_from_indexes(indexes: List[int], lang: str) -> Mnemonic:
    """Return the mnemonic from a list of word-list indexes.
    
    Return the mnemonic from a list of integer indexes into
    a given language word-list.
    """

    words = []
    wordlist = _wordlists.wordlist(lang)
    for i in indexes:
        word = wordlist[i]
        words.append(word)
    return ' '.join(words)

def indexes_from_mnemonic(mnemonic: Mnemonic, lang: str) -> List[int]:
    """Return the word-list indexes for a given mnemonic.
    
    Return the list of integer indexes into a language word-list
    for a given mnemonic.
    """

    words = mnemonic.split()
    wordlist = _wordlists.wordlist(lang)
    indexes = [wordlist.index(w) for w in words]
    return indexes

def entropy_from_indexes(indexes: List[int], lang: str) -> Entropy:
    """Return the entropy from a list of word-list indexes.
    
    Return the entropy from a list of integer indexes into
    a given language word-list.
    """

    n = _wordlists.language_length(lang)
    entropy = 0
    for i in indexes:
        entropy = entropy*n + i

    binentropy = bin(entropy)[2:]    # remove '0b'

    # do not lose leading zeros entropy
    bpw = _wordlists.bits_per_word(lang)
    bits = len(indexes)*bpw
    binentropy = binentropy.zfill(bits)

    return binentropy

def _seed_from_mnemonic(mnemonic: Mnemonic,
                        passphrase: str, prefix: str) -> bytes:
    hf_name = 'sha512'
    password = mnemonic.encode()
    salt = (prefix + passphrase).encode()
    iterations = 2048
    dksize = 64
    return pbkdf2_hmac(hf_name, password, salt, iterations, dksize)



from setuptools import setup, find_packages
import btclib

with open('README.md', 'r') as fh:
    longdescription = fh.read()

setup(
    name = btclib.name,
    version = btclib.__version__,
    url = 'http://github.com/dginst/btclib',
    license = btclib.__license__,
    author = 'Digital Gold Institute',
    description = 'A bitcoin cryptography library.',
    long_description = longdescription,
    long_description_content_type = 'text/markdown',
    packages = find_packages(exclude=['tests']),
    include_package_data = True,
    keywords = 'bitcoin cryptography elliptic-curves dsa schnorr RFC-6979 bip32 bip39 electrum base58',
    python_requires = '>=3.6',
    classifiers = [
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',

        'License :: OSI Approved :: MIT License',

        'Natural Language :: English',

        'Operating System :: OS Independent',

        'Topic :: Security :: Cryptography',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
